@@ -24,6 +24,10 @@
 #include <config.h>
 #include <atk/atkcomponent.h>
 #include <atk/atktext.h>
+#include <atk/atkhypertext.h>
+#include <atk/atkhyperlink.h>
+#include <glib/gi18n.h>
+#include <glib/gmacros.h>
 
 #include "gtkhtml.h"
 #include "htmlengine.h"
@@ -39,6 +43,7 @@
 #include "object.h"
 #include "html.h"
 #include "text.h"
+#include "hyperlink.h"
 
 static void html_a11y_text_class_init    (HTMLA11YTextClass *klass);
 static void html_a11y_text_init          (HTMLA11YText *a11y_text);
@@ -48,6 +53,8 @@ static void atk_text_interface_init      (AtkTextIface *iface);
 static void html_a11y_text_get_extents   (AtkComponent *component,
 					  gint *x, gint *y, gint *width, gint *height, AtkCoordType coord_type);
 static void html_a11y_text_get_size      (AtkComponent *component, gint *width, gint *height);
+static gboolean html_a11y_text_grab_focus (AtkComponent *comp);
+
 static gchar * html_a11y_text_get_text (AtkText *text, gint start_offset, gint end_offset);
 static gchar * html_a11y_text_get_text_after_offset (AtkText *text, gint offset, AtkTextBoundary boundary_type,
 						     gint *start_offset, gint *end_offset);
@@ -64,6 +71,19 @@ static gboolean html_a11y_text_remove_selection (AtkText *text, gint selection_n
 static gboolean html_a11y_text_set_selection (AtkText *text, gint selection_num, gint start_offset, gint end_offset);
 static gint html_a11y_text_get_caret_offset (AtkText *text);
 static gboolean html_a11y_text_set_caret_offset (AtkText *text, gint offset);
+static void html_a11y_text_get_character_extents (AtkText *text, gint offset,
+				gint *x, gint *y, gint *width, gint *height,
+				AtkCoordType coords);
+static gint html_a11y_text_get_offset_at_point (AtkText *text, gint x, gint y,
+					AtkCoordType coords);
+
+/* Hyperlink interface */
+static void atk_hyper_text_interface_init (AtkHypertextIface *iface);
+
+static AtkHyperlink * html_a11y_text_get_link (AtkHypertext *hypertext, gint link_index);
+static gint html_a11y_text_get_n_links (AtkHypertext *hypertext);
+static gint html_a11y_text_get_link_index (AtkHypertext *hypertext, gint char_index);
+
 
 /* Editable text interface. */
 static void 	atk_editable_text_interface_init      (AtkEditableTextIface *iface);
@@ -87,9 +107,45 @@ static void	html_a11y_text_paste_text	(AtkEditableText      *text,
 
 static AtkStateSet* html_a11y_text_ref_state_set	(AtkObject	*accessible);
 
-
 static AtkObjectClass *parent_class = NULL;
 
+static gint
+get_n_actions (AtkAction *action)
+{
+        return 1;
+}
+
+static G_CONST_RETURN gchar*
+action_get_name (AtkAction *action, gint      i)
+{
+        if (i == 0)
+                return _("grab focus");
+
+        return NULL;
+}
+
+static gboolean
+do_action (AtkAction * action, gint i)
+{
+        switch (i) {
+        case 0:
+		return html_a11y_text_grab_focus (ATK_COMPONENT (action));
+        default:
+                return FALSE;
+        }
+}
+
+static void
+atk_action_interface_init (AtkActionIface *iface)
+{
+        g_return_if_fail (iface != NULL);
+
+        iface->do_action = do_action;
+        iface->get_n_actions = get_n_actions;
+        iface->get_name = action_get_name;
+}
+
+
 GType
 html_a11y_text_get_type (void)
 {
@@ -128,15 +184,30 @@ html_a11y_text_get_type (void)
 			NULL
 		};
 
+		static const GInterfaceInfo atk_action_info = {
+			(GInterfaceInitFunc) atk_action_interface_init,
+			(GInterfaceFinalizeFunc) NULL,
+			NULL
+		};
+
+		static const GInterfaceInfo atk_hyper_text_info = {
+			(GInterfaceInitFunc) atk_hyper_text_interface_init,
+			(GInterfaceFinalizeFunc) NULL,
+			NULL
+		};
+
 		type = g_type_register_static (G_TYPE_HTML_A11Y, "HTMLA11YText", &tinfo, 0);
 		g_type_add_interface_static (type, ATK_TYPE_COMPONENT, &atk_component_info);
 		g_type_add_interface_static (type, ATK_TYPE_TEXT, &atk_text_info);
 		g_type_add_interface_static (type, ATK_TYPE_EDITABLE_TEXT, &atk_editable_text_info);
+		g_type_add_interface_static (type, ATK_TYPE_ACTION, &atk_action_info);
+		g_type_add_interface_static (type, ATK_TYPE_HYPERTEXT, &atk_hyper_text_info);
 	}
 
 	return type;
 }
 
+
 static void 
 atk_component_interface_init (AtkComponentIface *iface)
 {
@@ -144,6 +215,7 @@ atk_component_interface_init (AtkComponentIface *iface)
 
 	iface->get_extents = html_a11y_text_get_extents;
 	iface->get_size = html_a11y_text_get_size;
+	iface->grab_focus = html_a11y_text_grab_focus;
 }
 
 static void
@@ -164,6 +236,8 @@ atk_text_interface_init (AtkTextIface *iface)
 	iface->add_selection = html_a11y_text_add_selection;
 	iface->get_caret_offset = html_a11y_text_get_caret_offset;
 	iface->set_caret_offset = html_a11y_text_set_caret_offset;
+	iface->get_character_extents = html_a11y_text_get_character_extents;
+	iface->get_offset_at_point = html_a11y_text_get_offset_at_point;
 }
 
 static void
@@ -248,7 +322,8 @@ html_a11y_text_ref_state_set (AtkObject *accessible)
 		atk_state_set_add_state (state_set, ATK_STATE_EDITABLE);
 
 	atk_state_set_add_state (state_set, ATK_STATE_MULTI_LINE);
-	atk_state_set_add_state (state_set, ATK_STATE_MULTI_LINE);
+	atk_state_set_add_state (state_set, ATK_STATE_SENSITIVE);
+	atk_state_set_add_state (state_set, ATK_STATE_FOCUSABLE);
 
 	return state_set;
 }
@@ -257,44 +332,75 @@ html_a11y_text_ref_state_set (AtkObject *accessible)
  * AtkComponent interface
  */
 
-static void
-get_size (HTMLObject *obj, gint *width, gint *height)
-{
-	HTMLObject *last;
-
-	if (obj) {
-		gint ax, ay;
-
-		html_object_calc_abs_position (obj, &ax, &ay);
-		last = obj;
-		while (last->next && HTML_IS_TEXT_SLAVE (last->next))
-			last = last->next;
-		if (HTML_IS_TEXT_SLAVE (last)) {
-			gint lx, ly;
-			html_object_calc_abs_position (last, &lx, &ly);
-
-			*width = lx + last->width - ax;
-			*height = ly + last->descent - ay;
-		}
-	}
-}
 
 static void
 html_a11y_text_get_extents (AtkComponent *component, gint *x, gint *y, gint *width, gint *height, AtkCoordType coord_type)
 {
 	HTMLObject *obj = HTML_A11Y_HTML (component);
+	GtkHTMLA11Y *top_html_a11y;
+	HTMLEngine *top_engine;
+	gint min_x, min_y, max_x, max_y;
+	HTMLObject *next;
+	gint sx, sy;
 
-	html_a11y_get_extents (component, x, y, width, height, coord_type);
-	get_size (obj, width, height);
+	g_return_if_fail (obj);
+	top_html_a11y = html_a11y_get_top_gtkhtml_parent (HTML_A11Y (component));
+	g_return_if_fail (top_html_a11y);
+
+	if (obj->y <obj->ascent)
+		obj->y = obj->ascent;
+
+	atk_component_get_extents (ATK_COMPONENT (top_html_a11y), x, y, width, height, coord_type);
+
+	/* we need to get a max rect here */
+	html_object_calc_abs_position (obj, &min_x, &min_y);
+	max_x = min_x + obj->width;
+	max_y = min_y + obj->descent;
+	min_y -= obj->ascent;
+
+	next = obj->next;
+	while (next && HTML_IS_TEXT_SLAVE (next)) {
+		html_object_calc_abs_position (next, &sx, &sy);
+		min_x = MIN (min_x, sx);
+		min_y = MIN (min_y, sy - next->ascent);
+		max_x = MAX (max_x, sx + next->width);
+		max_y = MAX (max_y, sy + next->descent);
+
+		next = next->next;
+	}
+
+	*x += min_x;
+	*width = max_x - min_x;
+	*y += min_y;
+	*height = max_y - min_y;
+
+	/* scroll window */
+	top_engine = GTK_HTML_A11Y_GTKHTML (top_html_a11y)->engine;
+	*x -=  top_engine->x_offset;
+	*y -=  top_engine->y_offset;
 }
 
+
 static void
 html_a11y_text_get_size (AtkComponent *component, gint *width, gint *height)
 {
-	HTMLObject *obj = HTML_A11Y_HTML (component);
+	gint x, y;
 
-	html_a11y_get_size (component, width, height);
-	get_size (obj, width, height);
+	html_a11y_get_extents (component, &x, &y, width, height, ATK_XY_WINDOW);
+}
+
+static gboolean
+html_a11y_text_grab_focus (AtkComponent *comp)
+{
+	GtkHTML *html;
+
+	html = GTK_HTML_A11Y_GTKHTML (html_a11y_get_gtkhtml_parent (HTML_A11Y (comp)));
+	g_return_val_if_fail (html && html->engine && html_engine_get_editable (html->engine), FALSE);
+
+	html_engine_jump_to_object (html->engine, HTML_A11Y_HTML (comp), HTML_TEXT (HTML_A11Y_HTML (comp))->text_len);
+	g_signal_emit_by_name (html, "grab_focus");
+
+        return TRUE;
 }
 
 /*
@@ -375,7 +481,7 @@ html_a11y_text_get_text_after_offset (AtkText *text, gint offset, AtkTextBoundar
 
 	switch (boundary_type) {
 	case ATK_TEXT_BOUNDARY_LINE_START:
-		end_slave = html_text_get_slave_at_offset (HTML_OBJECT (to), offset);
+		end_slave = html_text_get_slave_at_offset (to, NULL, offset);
 		g_return_val_if_fail (end_slave, NULL);
 		start_slave = (HTMLTextSlave *) HTML_OBJECT (end_slave)->next;
 
@@ -394,7 +500,7 @@ html_a11y_text_get_text_after_offset (AtkText *text, gint offset, AtkTextBoundar
 		return html_a11y_text_get_text (text, *start_offset, *end_offset);
 
 	case ATK_TEXT_BOUNDARY_LINE_END:
-		start_slave = html_text_get_slave_at_offset (HTML_OBJECT (to), offset);
+		start_slave = html_text_get_slave_at_offset (to, NULL, offset);
 		g_return_val_if_fail (start_slave, NULL);
 
 		*start_offset = start_slave->posStart + start_slave->posLen;
@@ -425,7 +531,7 @@ html_a11y_text_get_text_at_offset (AtkText *text, gint offset, AtkTextBoundary b
 
 	switch (boundary_type) {
 	case ATK_TEXT_BOUNDARY_LINE_START:
-		start_slave = html_text_get_slave_at_offset (HTML_OBJECT (to), offset);
+		start_slave = html_text_get_slave_at_offset (to, NULL, offset);
 		g_return_val_if_fail (start_slave, NULL);
 		end_slave = (HTMLTextSlave *) HTML_OBJECT (start_slave)->next;
 
@@ -439,7 +545,7 @@ html_a11y_text_get_text_at_offset (AtkText *text, gint offset, AtkTextBoundary b
 		return html_a11y_text_get_text (text, *start_offset, *end_offset);
 
 	case ATK_TEXT_BOUNDARY_LINE_END:
-		end_slave = html_text_get_slave_at_offset (HTML_OBJECT (to), offset);
+		end_slave = html_text_get_slave_at_offset (to, NULL, offset);
 		g_return_val_if_fail (end_slave, NULL);
 		start_slave = (HTMLTextSlave *) HTML_OBJECT (end_slave)->prev;
 
@@ -482,7 +588,7 @@ html_a11y_text_get_text_before_offset (AtkText *text, gint offset, AtkTextBounda
 
 	switch (boundary_type) {
 	case ATK_TEXT_BOUNDARY_LINE_START:
-		end_slave = html_text_get_slave_at_offset (HTML_OBJECT (to), offset);
+		end_slave = html_text_get_slave_at_offset (to, NULL, offset);
 		g_return_val_if_fail (end_slave, NULL);
 		start_slave = (HTMLTextSlave *) HTML_OBJECT (end_slave)->prev;
 
@@ -495,7 +601,7 @@ html_a11y_text_get_text_before_offset (AtkText *text, gint offset, AtkTextBounda
 		return html_a11y_text_get_text (text, *start_offset, *end_offset);
 
 	case ATK_TEXT_BOUNDARY_LINE_END:
-		start_slave = html_text_get_slave_at_offset (HTML_OBJECT (to), offset);
+		start_slave = html_text_get_slave_at_offset (to, NULL, offset);
 		g_return_val_if_fail (start_slave, NULL);
 		end_slave = (HTMLTextSlave *) HTML_OBJECT (start_slave)->prev;
 
@@ -614,6 +720,83 @@ html_a11y_text_set_selection (AtkText *text, gint selection_num, gint start_offs
 
 */
 
+static gint
+html_a11y_text_get_offset_at_point (AtkText *text, gint x, gint y,
+					AtkCoordType coords)
+{
+	GtkHTML *top_html;
+	GtkHTMLA11Y *top_a11y;
+	HTMLEngine *top_e;
+	HTMLObject *obj, *return_obj;
+	gint offset = -1;
+	gint html_x, html_y, html_height, html_width;
+	gint text_x, text_y, text_height, text_width;
+
+	obj = HTML_A11Y_HTML (text);
+	g_return_val_if_fail(obj && html_object_is_text(obj), -1);
+
+	atk_component_get_extents (ATK_COMPONENT (text), &text_x, &text_y, &text_width, &text_height, coords);
+
+	/* check whether in range */
+	if (x < text_x || x > text_x + text_width
+		|| y < text_y || y > text_y + text_height)
+		return -1;
+
+	top_a11y = html_a11y_get_top_gtkhtml_parent (HTML_A11Y (text));
+	g_return_val_if_fail (top_a11y, -1);
+
+	top_html = GTK_HTML_A11Y_GTKHTML (top_a11y);
+	g_return_val_if_fail (top_html && GTK_IS_HTML(top_html) && top_html->engine, -1);
+	top_e = top_html->engine;
+
+	atk_component_get_extents (ATK_COMPONENT (top_a11y), &html_x, &html_y, &html_width, &html_height, coords);
+
+	x -= html_x;
+	y -= html_y;
+
+	return_obj = html_engine_get_object_at (top_e, x, y, &offset, FALSE);
+		
+	if (obj == return_obj)
+		return offset;
+	else
+		/*since the point is in range, need to return a valid value */ 
+		return 0;
+}
+
+static void
+html_a11y_text_get_character_extents (AtkText *text, gint offset,
+				gint *x, gint *y, gint *width, gint *height,
+				AtkCoordType coords)
+{
+	HTMLObject *obj;
+	GtkHTML *html;
+	HTMLEngine *e;
+	gint x1, x2, y1, y2;
+	GtkHTMLA11Y *a11y;
+
+	obj = HTML_A11Y_HTML (text);
+	g_return_if_fail(obj && html_object_is_text(obj));
+
+	a11y = html_a11y_get_top_gtkhtml_parent (HTML_A11Y (text));
+	g_return_if_fail (a11y);
+
+	html = GTK_HTML_A11Y_GTKHTML (a11y);
+	g_return_if_fail(html && GTK_IS_HTML(html) && html->engine);
+	e = html->engine;
+
+	atk_component_get_extents (ATK_COMPONENT (a11y), x, y, width, height, coords);
+	html_object_get_cursor (obj, e->painter, offset, &x1, &y1, &x2, &y2);
+
+	*x += x1;
+	*y += y1;
+	*height = y2 - y1;
+	/* a reasonable guess */
+	*width = *height / 2;
+
+	/* scroll window */
+	*x -=  e->x_offset;
+	*y -=  e->y_offset;
+}
  
 static void
 atk_editable_text_interface_init (AtkEditableTextIface *iface)
@@ -765,3 +948,40 @@ html_a11y_text_paste_text	(AtkEditableText *text,
 
         g_signal_emit_by_name(html, "grab_focus");
 }
+
+
+static void
+atk_hyper_text_interface_init (AtkHypertextIface *iface)
+{
+	g_return_if_fail (iface != NULL);
+
+	iface->get_link = html_a11y_text_get_link;
+	iface->get_n_links = html_a11y_text_get_n_links;
+	iface->get_link_index = html_a11y_text_get_link_index;
+}
+
+/*
+ * AtkHyperLink interface
+ */
+
+static AtkHyperlink *
+html_a11y_text_get_link (AtkHypertext *hypertext, gint link_index)
+{
+	return html_a11y_hyper_link_new (HTML_A11Y (hypertext), link_index);
+}
+
+static gint
+html_a11y_text_get_n_links (AtkHypertext *hypertext)
+{
+	HTMLText *text = HTML_TEXT (HTML_A11Y_HTML (hypertext));
+	return g_slist_length (text->links);
+}
+
+static gint
+html_a11y_text_get_link_index (AtkHypertext *hypertext, gint char_index)
+{
+	HTMLObject *obj = HTML_A11Y_HTML (hypertext);
+	Link *link = html_text_get_link_at_offset (HTML_TEXT (obj), char_index);
+
+	return link ? link->start_offset : -1;
+}
