@@ -26,6 +26,7 @@
 #include <string.h>
 #include <sys/types.h>
 #include <regex.h>
+#include <math.h>
 
 #include <pango/pango.h>
 
@@ -177,6 +178,7 @@ copy (HTMLObject *s,
 		csl->data = html_link_dup ((Link *) csl->data);
 
 	dest->pi = NULL;
+	dest->direction = src->direction;
 }
 
 /* static void
@@ -896,6 +898,8 @@ html_text_get_attr_list_list (PangoAttrList *get_attrs, PangoAttrList *attr_list
 				g_slist_free (l);
 			}
 		} while (pango_attr_iterator_next (iter));
+
+		pango_attr_iterator_destroy (iter);
 	}
 }
 
@@ -1185,7 +1189,7 @@ prepare_attrs (HTMLText *text, HTMLPainter *painter)
 		flow = HTML_CLUEFLOW (HTML_OBJECT (text)->parent);
 		
 	if (painter->widget && GTK_IS_HTML (painter->widget))
-		e = GTK_HTML (painter->widget)->engine;
+		e = html_object_engine (HTML_OBJECT (text), GTK_HTML (painter->widget)->engine);
 
 	if (flow && e)
 		html_text_add_cite_color (attrs, text, flow, e);
@@ -1195,20 +1199,27 @@ prepare_attrs (HTMLText *text, HTMLPainter *painter)
 		attr->start_index = 0;
 		attr->end_index = text->text_bytes;
 		pango_attr_list_insert (attrs, attr);
-		if (painter->font_manager.fix_size != painter->font_manager.var_size) {
-			attr = pango_attr_size_new (painter->font_manager.fix_size);
+		if (painter->font_manager.fix_size != painter->font_manager.var_size || fabs (painter->font_manager.magnification - 1.0) > 0.001) {
+			attr = pango_attr_size_new (painter->font_manager.fix_size*painter->font_manager.magnification);
+			attr->start_index = 0;
+			attr->end_index = text->text_bytes;
+			pango_attr_list_insert (attrs, attr);
+		}
+	} else {
+		if (fabs (painter->font_manager.magnification - 1.0) > 0.001) {
+			attr = pango_attr_size_new (painter->font_manager.var_size*painter->font_manager.magnification);
 			attr->start_index = 0;
 			attr->end_index = text->text_bytes;
 			pango_attr_list_insert (attrs, attr);
 		}
-	} else
 		pango_attr_list_splice (attrs, text->attr_list, 0, 0);
+	}
 
 	if (text->extra_attr_list)
 		pango_attr_list_splice (attrs, text->extra_attr_list, 0, 0);
 	if (!HTML_IS_PLAIN_PAINTER (painter)) {
 		if (flow && e)
-			html_text_change_attrs (attrs, html_clueflow_get_default_font_style (flow), GTK_HTML (painter->widget)->engine, 0, text->text_bytes, TRUE);
+			html_text_change_attrs (attrs, html_clueflow_get_default_font_style (flow), e, 0, text->text_bytes, TRUE);
 	}
 
 	if (text->links && e) {
@@ -1791,6 +1802,8 @@ save (HTMLObject *self, HTMLEngineSaveState *state)
 				html_text_free_attrs (attrs);
 			}
 		} while (pango_attr_iterator_next (iter));
+
+		pango_attr_iterator_destroy (iter);
 		g_slist_free (links);
 	}
 
@@ -2099,7 +2112,10 @@ static GList *
 remove_one (GList *list, GList *link)
 {
 	spell_error_destroy ((SpellError *) link->data);
-	return g_list_remove_link (list, link);
+	list = g_list_remove_link (list, link);
+	g_list_free (link);
+
+	return list;
 }
 
 static GList *
@@ -2447,39 +2463,45 @@ html_text_get_slave_at_offset (HTMLText *text, HTMLTextSlave *start, int offset)
 }
 
 static gboolean
-html_text_cursor_prev_slave (HTMLObject *slave, HTMLCursor *cursor)
+html_text_cursor_prev_slave (HTMLObject *slave, HTMLPainter *painter, HTMLCursor *cursor)
 {
-	HTMLObject *prev = HTML_OBJECT (slave)->prev;
 	int offset = cursor->offset;
 
-	if (prev && HTML_IS_TEXT_SLAVE (prev)) {
-		if (html_text_slave_cursor_tail (HTML_TEXT_SLAVE (prev), cursor)) {
-			cursor->position += cursor->offset - offset;
-			return TRUE;
+	while (slave->prev && HTML_IS_TEXT_SLAVE (slave->prev)) {
+		if (HTML_TEXT_SLAVE (slave->prev)->posLen) {
+			if (html_text_slave_cursor_tail (HTML_TEXT_SLAVE (slave->prev), cursor, painter)) {
+				cursor->position += cursor->offset - offset;
+				return TRUE;
+			} else
+				break;
 		}
+		slave = slave->prev;
 	}
 
 	return FALSE;
 }
 
 static gboolean
-html_text_cursor_next_slave (HTMLObject *slave, HTMLCursor *cursor)
+html_text_cursor_next_slave (HTMLObject *slave, HTMLPainter *painter, HTMLCursor *cursor)
 {
-	HTMLObject *next = slave->next;
 	int offset = cursor->offset;
 
-	if (next && HTML_IS_TEXT_SLAVE (next)) {
-		if (html_text_slave_cursor_head (HTML_TEXT_SLAVE (next), cursor)) {
-			cursor->position += cursor->offset - offset;
-			return TRUE;
+	while (slave->next && HTML_IS_TEXT_SLAVE (slave->next)) {
+		if (HTML_TEXT_SLAVE (slave->next)->posLen) {
+			if (html_text_slave_cursor_head (HTML_TEXT_SLAVE (slave->next), cursor, painter)) {
+				cursor->position += cursor->offset - offset;
+				return TRUE;
+			} else
+				break;
 		}
+		slave = slave->next;
 	}
 
 	return FALSE;
 }
 
 static gboolean
-html_text_cursor_right (HTMLObject *self, HTMLCursor *cursor)
+html_text_cursor_right (HTMLObject *self, HTMLPainter *painter, HTMLCursor *cursor)
 {
 	HTMLTextSlave *slave;
 
@@ -2489,14 +2511,14 @@ html_text_cursor_right (HTMLObject *self, HTMLCursor *cursor)
 	slave = html_text_get_slave_at_offset (HTML_TEXT (self), NULL, cursor->offset);
 
 	if (slave) {
-		if (html_text_slave_cursor_right (slave, cursor))
+		if (html_text_slave_cursor_right (slave, painter, cursor))
 			return TRUE;
 		else {
 			if (self->parent) {
 				if (html_object_get_direction (self->parent) == HTML_DIRECTION_RTL)
-					return html_text_cursor_prev_slave (HTML_OBJECT (slave), cursor);
+					return html_text_cursor_prev_slave (HTML_OBJECT (slave), painter, cursor);
 				else
-					return html_text_cursor_next_slave (HTML_OBJECT (slave), cursor);
+					return html_text_cursor_next_slave (HTML_OBJECT (slave), painter, cursor);
 			}
 		}
 	}
@@ -2505,7 +2527,7 @@ html_text_cursor_right (HTMLObject *self, HTMLCursor *cursor)
 }
 
 static gboolean
-html_text_cursor_left (HTMLObject *self, HTMLCursor *cursor)
+html_text_cursor_left (HTMLObject *self, HTMLPainter *painter, HTMLCursor *cursor)
 {
 	HTMLTextSlave *slave;
 
@@ -2515,14 +2537,14 @@ html_text_cursor_left (HTMLObject *self, HTMLCursor *cursor)
 	slave = html_text_get_slave_at_offset (HTML_TEXT (self), NULL, cursor->offset);
 
 	if (slave) {
-		if (html_text_slave_cursor_left (slave, cursor))
+		if (html_text_slave_cursor_left (slave, painter, cursor))
 			return TRUE;
 		else {
 			if (self->parent) {
 				if (html_object_get_direction (self->parent) == HTML_DIRECTION_RTL)
-					return html_text_cursor_next_slave (HTML_OBJECT (slave), cursor);
+					return html_text_cursor_next_slave (HTML_OBJECT (slave), painter, cursor);
 				else
-					return html_text_cursor_prev_slave (HTML_OBJECT (slave), cursor);
+					return html_text_cursor_prev_slave (HTML_OBJECT (slave), painter, cursor);
 			}
 		}
 	}
@@ -2531,12 +2553,12 @@ html_text_cursor_left (HTMLObject *self, HTMLCursor *cursor)
 }
 
 static int
-html_text_get_right_edge_offset (HTMLObject *o, int offset)
+html_text_get_right_edge_offset (HTMLObject *o, HTMLPainter *painter, int offset)
 {
 	HTMLTextSlave *slave = html_text_get_slave_at_offset (HTML_TEXT (o), NULL, offset);
 
 	if (slave) {
-		return html_text_slave_get_right_edge_offset (slave);
+		return html_text_slave_get_right_edge_offset (slave, painter);
 	} else {
 		g_warning ("getting right edge offset from text object without slave(s)");
 
@@ -2545,12 +2567,12 @@ html_text_get_right_edge_offset (HTMLObject *o, int offset)
 }
 
 static int
-html_text_get_left_edge_offset (HTMLObject *o, int offset)
+html_text_get_left_edge_offset (HTMLObject *o, HTMLPainter *painter, int offset)
 {
 	HTMLTextSlave *slave = html_text_get_slave_at_offset (HTML_TEXT (o), NULL, offset);
 
 	if (slave) {
-		return html_text_slave_get_left_edge_offset (slave);
+		return html_text_slave_get_left_edge_offset (slave, painter);
 	} else {
 		g_warning ("getting left edge offset from text object without slave(s)");
 
@@ -2907,7 +2929,7 @@ static HTMLMagicInsertMatch mim [] = {
 	{ "(news|telnet|nttp|file|http|ftp|https)://([-a-z0-9]+(:[-a-z0-9]+)?@)?[-a-z0-9.]+[-a-z0-9](:[0-9]*)?(/[-a-z0-9_$.+!*(),;:@%&=?/~#]*[^]'.}>\\) ,?!;:\"]?)?", NULL, NULL },
 	{ "www\\.[-a-z0-9.]+[-a-z0-9](:[0-9]*)?(/[-A-Za-z0-9_$.+!*(),;:@%&=?/~#]*[^]'.}>\\) ,?!;:\"]?)?", NULL, "http://" },
 	{ "ftp\\.[-a-z0-9.]+[-a-z0-9](:[0-9]*)?(/[-A-Za-z0-9_$.+!*(),;:@%&=?/~#]*[^]'.}>\\) ,?!;:\"]?)?", NULL, "ftp://" },
-	{ "[-_a-z0-9.\\+]+@[-_a-z0-9.]+", NULL, "mailto:" }
+	{ "[-_a-z0-9.'\\+]+@[-_a-z0-9.]+", NULL, "mailto:" }
 };
 
 #define MIM_N (sizeof (mim) / sizeof (mim [0]))
@@ -3686,3 +3708,16 @@ html_text_set_color (HTMLText *text, HTMLColor *color)
 {
 	html_text_set_color_in_range (text, color, 0, text->text_bytes);
 }
+
+HTMLDirection
+html_text_direction_pango_to_html (PangoDirection pdir)
+{
+	switch (pdir) {
+	case PANGO_DIRECTION_RTL:
+		return HTML_DIRECTION_RTL;
+	case PANGO_DIRECTION_LTR:
+		return HTML_DIRECTION_LTR;
+	default:
+		return HTML_DIRECTION_DERIVED;
+	}
+}
